# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.readthedocs.io/
# FOGLAMP_END

""" Plugin module to detect bad bearing """

import numpy as np
import pandas as pd
import os
import joblib
import json
import datetime
import copy
import logging
from foglamp.plugins.notificationRule.bad_bearing import wpt_features as wpt

from xgboost import XGBClassifier

from foglamp.common import logger

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2019 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"


_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Notification rule plugin to detect bad bearing',
        'type': 'string',
        'default': 'bad_bearing',
        'readonly': 'true'
    },
    'assetName': {
        'description': 'Name of asset',
        'type': 'string',
        'default': 'BearingVibration',
        'displayName': 'Asset name to monitor',
        'order': "1"
    },
    'samplingRate': {
        'description': 'Readings per sec',
        'type': 'integer',
        'default': '8000',
        'displayName': 'Sampling rate (per sec)',
        'order': "2"
    },
    'modelDirPath': {
        'description': 'Model dir path',
        'type': 'string',
        'default': os.path.join(os.path.dirname(__file__), "model"),
        'displayName': 'Model directory path',
        'order': "3"
    },
    'modelType': {
        'description': 'Model to use for bad bearing detection',
        'type': 'enumeration',
        'default': 'Neural network',
        'options': ['Neural network', 'Xgboost'],
        'displayName': 'Model',
        'order': '4'
    },
    'numAccChannels': {
        'description': 'Number of accelerometer channels to use',
        'type': 'integer',
        'default': '1',
        'displayName': 'Number of accelerometer channels to use',
        'minimum': '1',
        'maximum': '2',
        'order': '5'
    },
    'enableTpu': {
        'description': 'Connect the Coral Edge TPU and enable this',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Enable Edge TPU',
        'order': '6',
        'validity' : "modelType == 'Neural network'"
    }
}


_LOGGER = logger.setup(__name__, level=logging.WARN)

sampling_rate = 8000   # readings per sec
frame_size = sampling_rate
next_df = pd.DataFrame()
df = pd.DataFrame()
num_rows_processed = 0
num_input_rows = 0

try:
    from tflite_runtime.interpreter import Interpreter

    try:
        from tflite_runtime.interpreter import load_delegate
    except ImportError:
        _LOGGER.exception("Edge TPU Support is not found in Tensorflow installation ")
        pass
except:
    _LOGGER.exception("TFlite runtime installation not found")


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.info("plugin_info called")
    return {
        'name': 'bad_bearing',
        'version': '1.8.1',
        'mode': 'poll',
        'type': 'notificationRule',
        'interface': '1.0.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin.
    Args:
        config: JSON configuration document for the plugin configuration category
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    global sampling_rate, frame_size

    _LOGGER.info("plugin_init called: config={}".format(config))
    try:
        handle = copy.deepcopy(config)
        sampling_rate = int(handle['samplingRate']['value'])
        frame_size = sampling_rate

        if int(handle['numAccChannels']['value']) == 1:
            model_dir = os.path.join(handle['modelDirPath']['value'], '1-channel')
        else:
            model_dir = os.path.join(handle['modelDirPath']['value'], '2-channel')

        modelType = handle['modelType']['value']

        if modelType == 'Neural network':
            handle['enableTpu'] = handle['enableTpu']['value']

            if handle['enableTpu'] == 'true':
                handle['modelPath'] = os.path.join(model_dir, "bad_bearing_nn_edgetpu.tflite")
            else:
                handle['modelPath'] = os.path.join(model_dir, "bad_bearing_nn_float32.tflite")
        else:
            handle['modelPath'] = os.path.join(model_dir, "bad_bearing_xgb.model")

        handle['modelDataPath'] = os.path.join(model_dir, "model_data.npy")
        handle['pcaFilePath'] = os.path.join(model_dir, "kernel_PCA_32.dat")
        handle['scalerFilePath'] = os.path.join(model_dir, "min_max_scaler.dat")

        for file in ['modelPath', 'modelDataPath', 'pcaFilePath', 'scalerFilePath']:
            if not os.path.isfile(handle[file]):
                raise Exception("Unable to find file: {}".format(handle[file]))
            else:
                _LOGGER.info("Found file: {}".format(handle[file]))

        if modelType == 'Xgboost':
            handle['xgb_clf'] = joblib.load(handle['modelPath'])

        handle['pca'] = joblib.load(handle['pcaFilePath'])
        handle['scaler'] = joblib.load(handle['scalerFilePath'])

        model_data = np.load(handle['modelDataPath'], allow_pickle=True).item()

        # handle['fs'] = int(model_data['fs'])
        # handle['cutoff'] = int(model_data['cutoff'])

        if modelType == 'Neural network':
            try:
                t1 = datetime.datetime.now()
                if handle['enableTpu'] == 'true':
                    # load the Edge TPU Runtime
                    interpreter = Interpreter(model_path=handle['modelPath'],
                                              experimental_delegates=[load_delegate('libedgetpu.so.1.0')])
                else:
                    interpreter = Interpreter(model_path=handle['modelPath'])

                interpreter.allocate_tensors()
                handle['interpreter'] = interpreter
                t2 = datetime.datetime.now()
                delta = t2 - t1
                _LOGGER.info("Neural network model loaded in {} msec".format(int(delta.total_seconds() * 1000)))
            except Exception as e:
                handle['interpreter'] = None
                _LOGGER.info('Failed to create TFlite interpreter instance: {}{}'.
                             format(e, "   Check whether edge TPU is attached" if handle['enableTpu'] == 'true' else ""))

    except Exception as ex:
        _LOGGER.exception('Error in initializing bad_bearing plugin {}'.format(str(ex)))
        raise

    return handle


def plugin_triggers(handle):
    """ Returns assets to monitor for bearing vibration data
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        Returns the dict with asset name(s) under which bearing vibration data is available. This would be used to
        subscribe to such data from storage service
    """
    # d = {'triggers': [{'asset': handle['assetName']['value']}]}
    d = {'triggers': [{'asset': handle['assetName']['value'], 'All': 5}]}
    _LOGGER.info("plugin_triggers(): returning '{}'".format(d))
    return d


def plugin_eval(handle, data):
    """ Evaluates whether it is a bad bearing by running vibration data thru' ML model
    Args:
        handle: handle returned by the plugin initialisation call
        data: dict with the readings, containing vibration data for bearings
    Returns:
        Returns a bool indicating whether the bearing is bad
    """
    global next_df, df, num_rows_processed, num_input_rows, frame_size

    if handle['modelType']['value'] == 'Neural network' and handle['interpreter'] == None:
        # _LOGGER.error("Cannot process data since NN interpreter is not available")
        return False

    y_pred = False

    try:
        # t1 = datetime.datetime.now()
        data = json.loads(str(data))
        data = data[handle['assetName']['value']]
        print("data={}".format(data))
        if isinstance(data['channel1'], (int, float)):
            chunk = pd.DataFrame(data, index=[0])
        else:
            chunk = pd.DataFrame(data)

        num_input_rows += chunk.shape[0]

        df = chunk[:frame_size]
        _LOGGER.info("")

        if df.shape[0] < frame_size:
            _LOGGER.info("###### Got {} rows to process, NOT PROCESSING".format(chunk.shape[0]))
            return 0   # don't run ML inference if data length is not sufficient
        else:
            _LOGGER.info("++++++ Got {} rows to process, processing {} rows".format(chunk.shape[0], df.shape[0]))

        num_rows_processed += df.shape[0]

        interval_rows = int(frame_size*0.2)
        # _LOGGER.info("----> Rows={}, interval_rows={}, df.shape={}".format(df.shape[0], interval_rows, df.shape))

        features = wpt.parse_vibration_data(df, interval_rows, int(handle['numAccChannels']['value']))
        features = pd.DataFrame(features)
        # _LOGGER.info("features.shape={}".format(features.shape))
        # _LOGGER.info("handle['cutoff']={}, handle['fs']={}".format(handle['cutoff'], handle['fs']))

        features2 = features   # features.apply(lambda x: butter_lowpass_filtfilt(x, handle['cutoff'], handle['fs']))
        # _LOGGER.info("Butterworth low-pass filter applied")

        features3 = handle['scaler'].transform(features2)
        # _LOGGER.info("Min max scaler applied")
        features4 = handle['pca'].transform(features3)
        # _LOGGER.info("kernel PCA applied")

        modelType = handle['modelType']['value']
        if modelType == 'Xgboost':
            y_pred = handle['xgb_clf'].predict(features4)
            threshold = 0.9
        else:
            y_pred = nn_prediction(handle, features4)
            threshold = 0.8
            y_pred = y_pred.round()  # binarize probability to one of the 2 classes

        _LOGGER.info("y_pred={}".format(y_pred))
        _LOGGER.info("------ y_pred: sum={}, length={}, ratio={}, num_frames_processed={}, num_input_rows={}, so rv={}".
                     format(y_pred.sum(), len(y_pred),
                            (y_pred.sum() / len(y_pred)).round(2),
                            num_rows_processed // frame_size,
                            num_input_rows,
                            (y_pred.sum() / len(y_pred)) >= threshold))

        y_pred = True if ((y_pred.sum() / len(y_pred)) >= threshold) else False

        if df.shape[0] >= frame_size:
            df = next_df
            del next_df
            next_df = pd.DataFrame()

    except Exception as ex:
        _LOGGER.exception('Error in initializing bad_bearing plugin {}'.format(str(ex)))
        raise

    return y_pred   # 1 means bad bearing and also that this rule plugin wants to trigger notification


def plugin_reason(handle):
    """ Returns reason string for the last time plugin_eval returned True
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        Returns a simple dict with 'reason', 'timestamp' & 'asset' fields indicating the reason for last 'plugin_eval' call returning True
    """

    ts = str(datetime.datetime.now(datetime.timezone.utc))
    resp = { 'reason': 'triggered', 'timestamp': ts, 'asset': [ handle['assetName']['value'] ] }
    _LOGGER.info("plugin_reason(): response = {}".format(resp))
    return resp


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _LOGGER.info("Old config for bad_bearing plugin={} \n new config={}".format(handle, new_config))
    plugin_shutdown(handle)
    new_handle = plugin_init(new_config)
    return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """

    _LOGGER.info('bad_bearing plugin shut down.')
    handle['scaler'] = None


def nn_prediction(handle, features4):
    """ get prediction on features list using neural network
    Args:
        handle: handle returned by the plugin initialisation call
        features4: wavelet packet decomposition features selected by kernel PCA
    Returns:
        Returns prediction for each set of 32 PCs outputted by kernel PCA
    """
    features4 = features4.astype('float32')

    interpreter = handle['interpreter']
    # _LOGGER.info("interpreter.get_input_details()={}".format(interpreter.get_input_details()))
    # _LOGGER.info("interpreter.get_output_details()={}".format(interpreter.get_output_details()))
    output = interpreter.tensor(interpreter.get_output_details()[0]["index"])
    v = interpreter.get_input_details()[0]['quantization']
    # features3_copy = np.copy(features3)
    if handle['enableTpu'] == 'true':
        scale = v[0]
        zero_point = v[1]
        # _LOGGER.info("input side: scale={}, zero-point={}".format(scale, zero_point))
        features4 = np.round(features4 / scale + zero_point).astype('uint8')
    t1 = datetime.datetime.now()
    y_pred = []
    for r in range(features4.shape[0]):
        interpreter.set_tensor(interpreter.get_input_details()[0]['index'], np.expand_dims(features4[r], axis=0))
        interpreter.invoke()
        prediction = interpreter.get_tensor(interpreter.get_output_details()[0]['index'])[0][0]
        # _LOGGER.info("type(prediction)={}, prediction.shape={}, type(features3[r])={}, features3[r].shape={}"
        #                 .format(type(prediction), prediction.shape, type(features3[r]), features3[r].shape ))
        v = interpreter.get_output_details()[0]['quantization']
        scale = v[0]
        zero_point = v[1]
        # _LOGGER.info("output side: scale={}, zero-point={}".format(scale, zero_point))

        if handle['enableTpu'] == 'true':
            prediction2 = (prediction.astype('float32') - zero_point) * scale
            _LOGGER.debug("r={}, prediction={}, prediction2={:.3f}".format(r, prediction, prediction2))
            y_pred.append(prediction2)
        else:
            _LOGGER.debug("r={}, prediction={:.3f}".format(r, prediction))
            y_pred.append(prediction)

    y_pred = np.array(y_pred)
    t2 = datetime.datetime.now()
    delta = t2 - t1
    _LOGGER.info("Inference done in {} usec".format(int(delta.total_seconds() * 1000 * 1000)))

    return y_pred
